#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "../LnxDefs.h"
#include "../CommonData.h"
#include "../ConfigLabels.h"

#include "readuserparams.h"

/******************************************************************************
FUNCTION NAME:
	ReadUserParams

ABSTRACT:

PARAMETERS:
	IN  csPath -		String indicating the path to the driver
				configuration file
	OUT pAutoLogEnable -		
	OUT csAutoLogPath -			
	OUT csFirmwareFilename -	
	OUT pAutoLogFrequency -		
	OUT pAutoLogMaxSize -		
	OUT pAutoLogOverwrite -		

RETURN:
	The status of the operation.

DETAILS:
	
******************************************************************************/
NTSTATUS ReadUserParamsE(IN  const char *	csPath,
			 OUT unsigned long *	pAutoLogEnable,
			 OUT char *		csAutoLogPath,
			 OUT char *		csFirmwareFilename,
			 OUT unsigned long *	pAutoLogFrequency,
			 OUT unsigned long *	pAutoLogMaxSize,
			 OUT unsigned long *	pAutoLogOverwrite,
			 OUT TIG_USER_PARAMS	*pTigParams)
{
  char			LabelLine[257];
  FILE *		FileHandle;
  char			csFileName[512];
  //
  // This structure defines the labels and offsets for each
  // of the user configurable parameters.
  //
  typedef struct _USER_CFG
  {
    char *  	Label;
    void *  	Address;
    BOOL *  	pConfiged;
  } USER_CFG;

  const USER_CFG ConfigParams[] = {
    {AUTOLOG_PATH,	        csAutoLogPath,				NULL},
    {FIRMWARE_FILENAME,		csFirmwareFilename,			NULL},

    {RX_MAX_LATENCY,		&pTigParams->RxMaxLatency,		NULL},
    {RX_MIN_LATENCY,		&pTigParams->RxMinLatency,		NULL},
    {TX_MAX_LATENCY,		&pTigParams->TxMaxLatency,		NULL},
    {TX_MIN_LATENCY,		&pTigParams->TxMinLatency,		NULL},
    {RX_INTERRUPT_RATE,		&pTigParams->RxInterruptRate,		NULL},
    {TX_INTERRUPT_RATE,		&pTigParams->TxInterruptRate,		NULL},
    {RX_SPEED,			&pTigParams->RxSpeed,			NULL},
    {TX_SPEED,			&pTigParams->TxSpeed,			NULL},
    {RX_FIFO_SIZE,		&pTigParams->RxFifoSize,		NULL},
    {TX_FIFO_SIZE,		&pTigParams->TxFifoSize,		NULL},
    {RX_CHANNEL_SIZE,		&pTigParams->RxChannelSize,		&pTigParams->RxChannelSizeCfg},
    {RX_SEGMENTS,  		&pTigParams->RxSegments,		&pTigParams->RxSegmentsCfg},
    {TX_CHANNEL_SIZE,		&pTigParams->TxChannelSize,		&pTigParams->TxChannelSizeCfg},
    {TX_SEGMENTS,  		&pTigParams->TxSegments,		&pTigParams->TxSegmentsCfg},
    {ADSL_PSD_TEST_MODE,	&pTigParams->AdslPsdTestMode,		NULL},
    {LINE_PERSISTENT_ACT,	&pTigParams->LinePersistentAct,		NULL},
    {LINE_PERSISTENCE_TMR,	&pTigParams->LinePersistenceTmr,	NULL},
    {RX_BUF_ALLOC_LIMIT,	&pTigParams->RXBufAllocLimit,		NULL},
    {RX_BUF_RATIO_LIMIT,	&pTigParams->RXBufRatioLimit,		NULL},
    {RX_MAX_FRAME_SIZE,		&pTigParams->RxMaxFrameSize,		NULL},
    {TX_MAX_FRAME_SIZE,		&pTigParams->TxMaxFrameSize,		NULL},
    {MAX_TRANSMIT,		&pTigParams->MaxTransmit,		NULL},
    {RFC_ENCAPSULATION_MODE,	&pTigParams->RfcEncapsulationMode,	NULL},
    {CELL_BURST_SIZE,		&pTigParams->CellBurstSize,		NULL},
    {PEAK_CELL_RATE,		&pTigParams->PeakCellRate,		NULL},
    {CONTROLLER_LOG_ENABLE,	&pTigParams->CdALDiagControllerLogEnable,	NULL},
    {CAPABILITIES_G922,		&pTigParams->BdADSLLocalG922Cap,	NULL},
    {CAPABILITIES_G922_ANNEX,	&pTigParams->BdADSLLocalG922AnnexCap,	NULL},
    {LINE_AUTO_ACTIVATION,	&pTigParams->LineAutoActivation,	NULL},
    {AUTO_SENSE_HANDSHAKE,	&pTigParams->AutoSenseHandshake,	NULL},
    {AUTO_SENSE_WIRES_FIRST,	&pTigParams->AutoSenseWiresFirst,	NULL},
    {AUTO_WIRING_SELECTION,	&pTigParams->AutoWiringSelection,	NULL},
    {AUTO_WIRING_RELAY_DELAY,	&pTigParams->AutoWiringRelayDelay,	NULL},
    {AUTO_WIRING_RELAY_ENRG,	&pTigParams->AutoWiringRelayEnrg,	NULL},
    {AUTO_WIRING_NUM_COMBOS,	&pTigParams->AutoWiringNumCombos,	NULL},
    {AUTO_WIRING_COMBO_GPIO,	&pTigParams->AutoWiringComboGPIO,	NULL},
    {AUTO_WIRING_ORDER,		&pTigParams->AutoWiringOrder,		NULL},
    {ADSL_HEAD_END,		&pTigParams->AdslHeadEnd,		NULL},
    {ADSL_HEAD_END_ENVIRONMENT,	&pTigParams->AdslHeadEndEnvironment,	NULL},
    {VENDOR_NEAR_ID,		&pTigParams->VendorNearId,		NULL},
    {DEBUG_FLAG,		&pTigParams->DebugFlag,			NULL},
    {API_HWIF_MODE,		&pTigParams->ApiHwIfMode,		NULL},
    {OVERRIDE_MAC_ADDRESS,	&pTigParams->OverrideMacAddress,	NULL},
    {MAC_ADDRESS,		&pTigParams->MACAddress,		NULL},
    {AUTOLOG_ENABLED,		&pTigParams->AutoLogEnabled,		NULL},
    {AUTOLOG_MAX_FILE_SIZE,	&pTigParams->AutoLogMaxFileSize,	NULL},
    {AUTOLOG_UPDATE_FREQUENCY,	&pTigParams->AutoLogUpdateFrequency,	NULL},
    {AUTOLOG_OVERWRITE,		&pTigParams->AutoLogOverwrite,		NULL},
    {ARM_HW_ID,			&pTigParams->ARMHwId,			NULL},
    {ADSL_HW_ID,		&pTigParams->ADSLHwId,			NULL},
    {ARM_FUNCTION_IDENTIFIER,	&pTigParams->ARMFunctionIdentifier,	NULL},
    {ADSL_FUNCTION_IDENTIFIER,	&pTigParams->ADSLFunctionIdentifier,	NULL},
    {HW_DEBUG_FLAG,		&pTigParams->HwDebugFlag,		NULL},
    {PHYSICAL_DRIVER_TYPE,	&pTigParams->PhysicalDriverType,	NULL},
    {GPIO_ADSL_FUNCTION_MASK,	&pTigParams->GpioADSLFunctionMask,	NULL},
    {GPIO_ARM_FUNCTION_MASK,	&pTigParams->GpioARMFunctionMask,	NULL},
    {GPIO_DIRECTION_MASK,	&pTigParams->GpioDirectionMask,		NULL},
    {GPIO_EDGE_MASK,		&pTigParams->GpioEdgeMask,		NULL},
  };

  // set the entries to default values
  memset(pTigParams, 0, sizeof(*pTigParams));
  strcpy(csFirmwareFilename, "Default.hex");

  // open the user configuration file
  strcpy(csFileName, csPath);
  strcat(csFileName, "/");
  strcat(csFileName, CONFIGURATION_FILENAME);
  FileHandle = fopen(csFileName, "r");
  if (FileHandle <= 0)
  {
    printf("***ERROR - User params file (%s) not found or unable to open!\n",
	    csFileName);
    return STATUS_FAILURE;
  }

  // loop while there are still lines in the file, read a line and quit on eof
  while (fgets(LabelLine, sizeof(LabelLine), FileHandle))
  {
    int		CharCnt;
    char	*pValue;
    char	*pLabel = LabelLine;

    // skip leading spaces
    while ((*pLabel == ' ')  ||
	   (*pLabel == '\t') ||
	   (*pLabel == 0x0a) ||
	   (*pLabel == 0x0d))
      pLabel++;

    // skip it if its a comment. a comment starts
    // with any non-alpha character
    if (!pLabel || !isalpha(*pLabel))
      continue;

    // find the value
    pValue = pLabel;
    for (CharCnt = 0; CharCnt < 256; CharCnt++)
    {
      if (!isalpha(*pValue))
      {
	*pValue = 0;
	pValue++;
	break;
      }

      pValue++;
    }

    // find the label and set the value
    {
      long		cnt;

      // for each entry in table
      for (cnt=0; cnt < sizeof(ConfigParams)/sizeof(ConfigParams[0]); cnt++)
      {
	// does this line match table entry?
	if (!strcmp(pLabel, ConfigParams[cnt].Label))
	{
	  long	value;
	  void *pStorage;	// location converted value is saved to
	  BOOL *pConfiged;	// location of indicator that value was found
	  void *pConv;		// location to convert value to
	  char	CfgFormat[] = "%ld";	// default format
	  char	ScanFormat[] = "%ld";	// default format

	  // skip intervening whitespace
	  while ((*pValue == ' ') || (*pValue == '\t'))
	    pValue++;

	  // is there a format string?
	  if (*pValue == '%')
	  {
	    // skip over '%'
	    pValue++;

	    // does the format string have a length modifier
	    switch (*pValue)
	    {
	    case 'b':
	    case 'h':
	    case 'l':
	      CfgFormat[1] = *pValue;
	      pValue++;
	      break;
	    default:
	      CfgFormat[1] = 'l';
	    }

	    // default converstion location is temp var
	    pConv = &value;

	    // which conversion char is specified
	    switch (*pValue)
	    {
	    case 'x':
	    case 'd':
	      CfgFormat[2] = *pValue;
	      ScanFormat[2] = *pValue;
	      pValue++;
	      break;

	    case 's':
	      CfgFormat[1] = *pValue;
	      ScanFormat[1] = *pValue;

	      CfgFormat[2] = 0;
	      ScanFormat[2] = 0;

	      // convert strings rigth into storage location
	      pConv = ConfigParams[cnt].Address;

	      pValue++;
	      break;

	    default:
	      CfgFormat[2] = 'd';
	      ScanFormat[2] = 'd';
	      break;
	    }
	  }
	  else
	    pConv = &value;

	  // Get ptr to where to store value
	  pStorage = ConfigParams[cnt].Address;

	  // convert each value
	  while (sscanf(pValue, ScanFormat, pConv) != EOF)
	  {
	    // transfer according to type
	    switch (CfgFormat[1])
	    {
	    case 's':
		// nothing to do, value was scanned to storage location
		break;

	    case 'b':
	      {
		unsigned char *pbVal = (unsigned char *) pStorage;

		* pbVal = (unsigned char) value;
		pbVal++;
		pStorage = (ULONG *) pbVal;
		break;
	      }

	    case 'h':
	      {
		unsigned short int *phVal = (unsigned short int *) pStorage;

		* phVal = (unsigned short int) value;
		phVal++;
		pStorage = (ULONG *) phVal;
		break;
	      }

	    case 'l':
	    default:
	      {
		unsigned long int *plVal = (unsigned long int *) pStorage;

		* plVal = (unsigned long int) value;
		plVal++;
		pStorage = (ULONG *) plVal;
		break;
	      }
	    }

	    // skip to next field
	    while (*pValue != ',')
	    {
	      if (*pValue == 0)
	      {
		pValue--;
		break;
	      }
	      pValue++;
	    }
	    pValue++;
	  }

	  // Get ptr to where to store indicator that it was found
	  pConfiged = ConfigParams[cnt].pConfiged;
	  if (pConfiged)
	    *pConfiged = TRUE;

	  // exit label search loop
	  break;
	}
      }
    }
  }

  // close the file
  fclose(FileHandle);

  *pAutoLogEnable = pTigParams->AutoLogEnabled;
  *pAutoLogFrequency = pTigParams->AutoLogUpdateFrequency;
  *pAutoLogMaxSize = pTigParams->AutoLogMaxFileSize;
  *pAutoLogOverwrite = pTigParams->AutoLogOverwrite;

  return STATUS_SUCCESS;
}
