#ifndef _KTHREAD_H
#define _KTHREAD_H

/*
  This source appears to be from:
  http://www.scs.ch/~frey/linux/kernelthreads.html

  Author: rey@scs.ch
*/

#define NEWTHREAD 1

#include "linuxmodver.h"

#include <linux/sched.h>
#if defined(INIT_SIGHAND) || (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
# define SIGMASK_LOCK sighand->siglock
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,26))
# define SIGMASK_LOCK sig->siglock
#else
# define SIGMASK_LOCK sigmask_lock
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/workqueue.h>
#else
#include <linux/tqueue.h>
#endif
#include <linux/wait.h>

#include <asm/semaphore.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
#define KTHREAD_R void
#else
#define KTHREAD_R int
typedef struct wait_queue * wait_queue_head_t;
#define init_MUTEX_LOCKED(x) *(x) = MUTEX_LOCKED
#endif

// 2.4.9 introduced complete_and_exit
#if !NEWTHREAD || (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,9))
#define COMPLETION semaphore
#define DECLARE_COMPLETION(x) DECLARE_MUTEX_LOCKED(x)
#define init_completion(x) init_MUTEX_LOCKED(x)
#define complete(x) up(x)
#if 0
#define complete_and_exit(x, y) up_and_exit((x), (y))
#else
#define complete_and_exit(x, y) up(x)
#endif
#define wait_for_completion(x) down(x)
#else
#include <linux/completion.h>
#define COMPLETION completion
#endif

/* a structure to store all information we need for our thread */
typedef struct kthread_struct
{
  /* private data */

  /* pid of thread */
  int pid;

  /* Task queue needed to launch thread */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
  struct work_struct tq;
#else
  struct tq_struct tq;
#endif

  /* function to be started as thread */
  void (*function) (struct kthread_struct *kthread);

  /* semaphore needed on start and creation of thread. */
  struct COMPLETION startstop_sem;

  /* public data */

  /* queue thread is waiting on. Gets initialized by
     init_kthread, can be used by thread itself.
  */
  wait_queue_head_t queue;

  /* flag to tell thread whether to die or not.
     When the thread receives a signal, it must check
     the value of terminate and call exit_kthread and terminate
     if set.
  */
  int terminate;

  /* additional data to pass to kernel thread */
  int which;
  void *arg;
} kthread_t;

/* prototypes */

/* start new kthread (called by creator) */
void start_kthread(KTHREAD_R (*func)(kthread_t *), kthread_t *kthread);

/* stop a running thread (called by "killer") */
void stop_kthread(kthread_t *kthread);

/* setup thread environment (called by new thread) */
void init_kthread(kthread_t *kthread, const char *name, int boost);

/* cleanup thread environment (called by thread upon receiving termination signal) */
void exit_kthread(kthread_t *kthread);
#endif
